/*******************************************************************************
 *
 * Driver for SSD1306 Display written as CFunctions
 *
 * Author: Peter Mather 2015 with acknowledgements to Peter Carnegie & Geoff Graham
 * 
 *
 * This CFunction MUST be compiled with Optimization Level 1, -O1
 * -O2,-O3,-Os will compile successfully, but generate exceptions at runtime.
 *
 * When Generating the CFunction, use MERGE CFunction mode, and name the CFunction
 * SSD1306
 *
 * Entry point is function long long main(long long *address //i2c only
 *  *                                     long long *DC,
 *                                        long long *RST
 *                                        long long *CS //SPI only
 *                                        long long *size //zero for 0.96", non-zero for 1.3"
 *                                        long long *orientation)
 *
 * V1.0     2015-10-15 Peter Mather
 * V2.0     2016-2-29  V5.0 compatibility, I2C bitbang impemlentation
 ******************************************************************************/
#include <stdarg.h>

#define Version 200     //Version 2.00
#define _SUPPRESS_PLIB_WARNING                                      // required for XC1.33  Later compiler versions will need PLIB to be installed
#include <plib.h>                                                   // the pre Harmony peripheral libraries
#include "../cfunctions.h"
#define i2c //define for compilation for I2c, comment out for SPI
#define LANDSCAPE       1
#define PORTRAIT        2
#define RLANDSCAPE      3
#define RPORTRAIT       4

#define screenwidth 128
#define screenheight Option->pins[0]
#define screenrows screenheight/8
#define update screenwidth*screenrows+1 //location for status to force screen write
#define magic 0
// SPI pin numbers and registers
#define SPI_INP_PIN         (HAS_44PINS ? 41 : 14)
#define SPI_OUT_PIN         (HAS_44PINS ? 20 :  3)
#define SPI_CLK_PIN         (HAS_44PINS ? 14 : 25)
#define SPI_PPS_OPEN        PPSInput(2, SDI1, RPB5); PPSOutput(2, RPA1, SDO1)
#define SPI_PPS_CLOSE       PPSOutput(2, RPA1, NULL)
#define SPICON *(volatile unsigned int *)(0xbf805800)               //SPI status register
#define SPISTAT *(volatile unsigned int *)(0xbf805810)              //SPI status register
#define SPIBUF *(volatile unsigned int *)(0xbf805820)               //SPI output buffer
#define SPIBRG *(volatile unsigned int *)(0xbf805830)               //SPI output buffer
#define SPICON2 *(volatile unsigned int *)(0xbf805840)              //SPI status register

#define SPIsend(a) {int j;SPIBUF=a; while((SPISTAT & 0x80)==0); j=SPIBUF;}
#ifdef i2c
#define dat Option->LCD_Reset
#define clk Option->LCD_CD
#define SetHi(p) PinSetBit(p, TRISSET)
#define SetLo(p) PinSetBit(p, TRISCLR)


// delay for a certain number of core timer ticks (one tick every 50nS at 40MHz CPU speed, 100nS at 20MHz, etc)
inline void Delay(unsigned int ticks) {
    unsigned int current_ticks;
    asm volatile("mfc0   %0, $9" : "=r"(current_ticks));            // get the core timer ticks
    ticks += current_ticks;                                         // this is what we want to count up to
     do {
         asm volatile("mfc0   %0, $9" : "=r"(current_ticks));       // get the core timer ticks
     } while(current_ticks < ticks);
}

// bit banged I2C Start condition, data line goes low when clock is high
void BBStart(int speed) {
    SetHi(dat); SetHi(clk);
    Delay(speed);
    SetLo(dat);
    Delay(speed);
    SetLo(clk);
    Delay(speed);
}


// bit banged I2C Stop condition, clock goes high when data is low
void BBStop (int speed) {
    SetLo(clk);
    SetLo(dat);
    Delay(speed);
    SetHi(clk);
    Delay(speed);
    SetHi(dat);
    Delay(speed);
}


// bit banged I2C write a bit
inline void BBWriteBit(int speed, int timeout, int bit) {
    int i = 0;
    if(bit)
        SetHi(dat);
    else
        SetLo(dat);
    SetHi(clk);
    asm volatile("mtc0   %0, $9": "+r"(i));                         // set the core timer to zero
    while(!PinRead(clk)) {                                          // Wait for clock to go high - clock stretching
         asm volatile("mfc0   %0, $9" : "=r"(i));                   // get the core timer ticks
         if(i >= timeout) 
         return;                                                // return nothing if there is a timeout
    }
    Delay(speed); SetLo(clk);
}


// bit banged I2C read a bit
inline int BBReadBit(int speed, int timeout) {
    int i = 0;
    SetHi(dat); Delay(speed);
    SetHi(clk); Delay(speed);
    asm volatile("mtc0   %0, $9": "+r"(i));                         // set the core timer to zero
    while(!PinRead(clk)) {                                          // Wait for clock to go high - clock stretching
         asm volatile("mfc0   %0, $9" : "=r"(i));                   // get the core timer ticks
         if(i >= timeout)

             return 0;                                              // return nothing if there is a timeout
    }
    i = PinRead(dat);
    SetLo(clk);
    return i;
}


// bit banged I2C Write, returns true if ACK received
int BBWriteByte(int speed, int timeout, int data) {
    int i;

    for(i = 0; i < 8; i++) {
        BBWriteBit(speed, timeout, (data & 0x80));
        data  <<= 1;
    }
    return !BBReadBit(speed, timeout);                    // read the ACK bit and return true is ACK received
}



int writei2c (int nbr, int addr, unsigned char *data) {
   unsigned  int i=nbr, speed= (CurrentCpuSpeed/1600000);
   unsigned int timeout = (CurrentCpuSpeed/2000) * 5000;           // calculate the number of core timer ticks for the timeout (5mS)
        BBStart(speed);
        if(!BBWriteByte(speed, timeout, (addr << 1)))
            return 0;
        while(i--)
            if(!BBWriteByte(speed, timeout, *data++))
                return 0;
        BBStop(speed);
        return 1;
    }
void oled_write_command(unsigned char data){
    char sc[2];
    sc[0]=0;
    sc[1]=data;
    writei2c(2,Option->pins[1],sc);
}
#else
void oled_write_command(unsigned char data){
    PinSetBit(Option->LCD_CD, LATCLR);
    PinSetBit(Option->LCD_CS, LATCLR);
    SPIsend(data);
    PinSetBit(Option->LCD_CS, LATSET);
}

#endif
void OLED_setxy(int x,int y){
  oled_write_command(0xB0+y)        ; //set page address 
  oled_write_command(0x10 | ((x>>4) & 0x0F))  ; //set high col address 
  oled_write_command(x & 0x0f)     ; //set low col address 
}
#ifdef i2c
void updatedisplay(void){
    char* p=(void *)(unsigned int)CFuncRam[0];
    if(p[update]){
    char *q;
    char saveit;
         int n;
        p[update]=0;
        for(n=0;n<screenrows;n++){
            OLED_setxy(0,n);
            q=p+n*screenwidth;
            saveit=q[0];
            q[0]=0x40;
            writei2c(screenwidth+1,Option->pins[1],q);
            q[0]=saveit;
        } 
    }
}
#else
void updatedisplay(void){
    char* p=(void *)(unsigned int)CFuncRam[0];
    if(p[update]){
        int m,n;
        unsigned int consave=0,brgsave=0,con2save;
        p[update]=0;
        brgsave=SPIBRG; //save any user SPI setup
        consave=SPICON;
        con2save=SPICON2;
        SPICON=0x8060;
        SPIBRG=2;
        SPICON2=0xC00;
        for(n=0;n<screenrows;n++){
            OLED_setxy(0,n);
            PinSetBit(Option->LCD_CS, LATCLR);
            PinSetBit(Option->LCD_CD, LATSET);
            for(m=n*screenwidth;m<(n+1)*screenwidth;m++){
                SPIsend(p[m+1]);
            }
            PinSetBit(Option->LCD_CS, LATCLR);
        } 
        SPIBRG=brgsave;  //restore user (or my) setup
        SPICON=consave;
        SPICON2=con2save;
    }
}
#endif

long long DrawRectangleMEM(int x1, int y1, int x2, int y2, int c){
    unsigned char* p=(void *)(unsigned int)CFuncRam[0];
    if(p[magic]!=0xF5){
        p = GetMemory(screenwidth*screenrows+1024);
    //Save the address for other CFunctions
        CFuncRam[0]=(unsigned int)(p);
        p[magic]=0xF5;
    }
    int i,j,loc,t;
    unsigned char mask;
    if(Option->DISPLAY_ORIENTATION==PORTRAIT){
        t=x1;
        x1=screenwidth-y2-1;
        y2=t;
        t=x2;
        x2=screenwidth-y1-1;
        y1=t;
    }
     if(Option->DISPLAY_ORIENTATION==RLANDSCAPE){
        x1=screenwidth-x1-1;
        x2=screenwidth-x2-1;
        y1=screenheight-y1-1;
        y2=screenheight-y2-1;
    }
   if(Option->DISPLAY_ORIENTATION==RPORTRAIT){
        t=y1;
        y1=screenheight-x1-1;
        x1=t;
        t=y2;
        y2=screenheight-x2-1;
        x2=t;
    }
    if(x2 <= x1) { t = x1; x1 = x2; x2 = t; }
    if(y2 <= y1) { t = y1; y1 = y2; y2 = t; }
    if(x1 < 0) x1 = 0; if(x1 >= screenwidth) x1 = screenwidth - 1;
    if(x2 < 0) x2 = 0; if(x2 >= screenwidth) x2 = screenwidth - 1;
    if(y1 < 0) y1 = 0; if(y1 >= screenheight) y1 = screenheight - 1;
    if(y2 < 0) y2 = 0; if(y2 >= screenheight) y2 = screenheight - 1;
        
     for(i=x1;i<=x2;i++){
        for(j=y1;j<=y2;j++){
           loc=1+i+(j/8)*screenwidth; //get the byte address for this bit
           mask=1<<(j % 8); //get the bit position for this bit
           if(c){
               p[loc]|=mask;
           } else {
               p[loc]&=(~mask);
           }
        }
    }
    p[update]=1;
}
void DrawBitmapMEM(int x1, int y1, int width, int height, int scale, int fc, int bc, unsigned char *bitmap){
    int i, j, k, m, x, y,t, loc;
    unsigned char omask, amask;
    unsigned char* p=(void *)(unsigned int)CFuncRam[0];
    if(p[magic]!=0xF5){
        p = GetMemory(screenwidth*screenrows+1024);
    //Save the address for other CFunctions
        CFuncRam[0]=(unsigned int)(p);
        p[magic]=0xF5;
    }
    for(i = 0; i < height; i++) {                                   // step thru the font scan line by line
        for(j = 0; j < scale; j++) {                                // repeat lines to scale the font
            for(k = 0; k < width; k++) {                            // step through each bit in a scan line
                for(m = 0; m < scale; m++) {                        // repeat pixels to scale in the x axis
                    x=x1 + k * scale + m ;
                    y=y1 + i * scale + j ;
                    if(Option->DISPLAY_ORIENTATION==PORTRAIT){
                        t=x;
                        x=VRes-y-1;
                        y=t;
                    }
                    if(Option->DISPLAY_ORIENTATION==RLANDSCAPE){
                        x=HRes-x-1;
                        y=VRes-y-1;
                    }
                    if(Option->DISPLAY_ORIENTATION==RPORTRAIT){
                        t=y;
                        y=HRes-x-1;
                        x=t;
                    }
                    loc=1+x+(y/8)*screenwidth; //get the byte address for this bit
                    omask=1<<(y % 8); //get the bit position for this bit
                    amask=~omask;
                    if(x >= 0 && x < screenwidth && y >= 0 && y < screenheight) {  // if the coordinates are valid
                        if((bitmap[((i * width) + k)/8] >> (((height * width) - ((i * width) + k) - 1) %8)) & 1) {
                            if(fc){
                                p[loc]|=omask;
                             } else {
                                p[loc]&=amask;
                             }
                       } else {
                            if(bc>0){
                                p[loc]|=omask;
                            } else if(bc==0) {
                                p[loc]&=amask;
                            }
                        }
                   }
                }
            }
        }
    }
    p[update]=1;
}
__attribute__((noinline)) void getFPC(void *a, void *b, volatile unsigned int *c) 
     { 
         *c = (unsigned int) (__builtin_return_address (0) - (b -a)) ;      
     } 
void pstring(const char *s){
    volatile unsigned int libAddr ; 
    getFPC(NULL,&&getFPCLab,&libAddr) ; // warning can be ignored, stupid editor 
    getFPCLab: { } 
    unsigned char  * testData    = (unsigned char *)((void *)s + libAddr );
    MMPrintString(testData);
}
//CFunction Driver_SSD1306
#ifdef i2c
long long main(long long *address, long long *CD, long long *RST, long long *size, long long *orientation){
    // for i2c RST is used for SDA, CD is used for SCK
#else
long long main(long long *CD, long long *RST, long long *CS,long long *size, long long *orientation){
#endif
    volatile unsigned int libAddr ; 
    getFPC(NULL,&&getFPCLab,&libAddr) ; // warning can be ignored, stupid editor 
    getFPCLab: { } 
    screenheight=*size;
    int HorizontalRes=screenwidth,VerticalRes=screenheight;
    Option->LCD_Reset=*RST;
    Option->LCD_CD=*CD;
    Option->DISPLAY_TYPE=20;
#ifdef i2c
    Option->pins[1]=*address;
#endif
#ifndef i2c
    unsigned int consave=0,brgsave=0,con2save=0;
    Option->LCD_CS=*CS;
#endif
    Option->DISPLAY_ORIENTATION=*orientation;
    //Get some persistent memory
    unsigned char * p = GetMemory(screenwidth*screenrows+1024);
    //Save the address for other CFunctions
    CFuncRam[0]=(unsigned int)(p);
    p[magic]=0xF5;
    PinSetBit(Option->LCD_Reset, LATCLR);
    ExtCfg(Option->LCD_Reset,EXT_DIG_OUT,0);ExtCfg(Option->LCD_Reset,EXT_BOOT_RESERVED,0);
    PinSetBit(Option->LCD_CD, LATCLR);
    ExtCfg(Option->LCD_CD,EXT_DIG_OUT,0);ExtCfg(Option->LCD_CD,EXT_BOOT_RESERVED,0);
    
#ifndef i2c
    ExtCfg(Option->LCD_CS,EXT_DIG_OUT,0);ExtCfg(Option->LCD_CS,EXT_BOOT_RESERVED,0);
    PinSetBit(Option->LCD_CS, LATSET);
    if(ExtCurrentConfig[SPI_OUT_PIN] == EXT_RESERVED) { //already open
        brgsave=SPIBRG;
        consave=SPICON;
        con2save=SPICON2;
    }

    ExtCfg(SPI_OUT_PIN, EXT_DIG_OUT, 0); ExtCfg(SPI_OUT_PIN, EXT_BOOT_RESERVED, 0);
    ExtCfg(SPI_INP_PIN, EXT_DIG_IN, 0); ExtCfg(SPI_INP_PIN, EXT_BOOT_RESERVED, 0);
    ExtCfg(SPI_CLK_PIN, EXT_DIG_OUT, 0); ExtCfg(SPI_CLK_PIN, EXT_BOOT_RESERVED, 0);
    SPI_PPS_OPEN; 
    SPICON=0x8060;
    SPIBRG=2;
    SPICON2=0xC00;// this is defined in IOPorts.h
    if(!brgsave){ //save my settings
        brgsave=SPIBRG;
        consave=SPICON;
        con2save=SPICON2;
    }    



   //Reset the SSD1306
    PinSetBit(Option->LCD_Reset,LATSET);
    uSec(10000);
    PinSetBit(Option->LCD_Reset,LATCLR);
    uSec(10000);
    PinSetBit(Option->LCD_Reset,LATSET);
    uSec(10000);
#endif
    oled_write_command(0xAE); //DISPLAYOFF) 
    oled_write_command(0xD5); //DISPLAYCLOCKDIV 
    oled_write_command(0x80); //the suggested ratio 0x80 
    oled_write_command(0xA8); //MULTIPLEX 
    if(screenheight==64)oled_write_command(0x3F); 
    else oled_write_command(0x1F);
    oled_write_command(0xD3); //DISPLAYOFFSET 
    oled_write_command(0x0); //no offset 
    oled_write_command(0x40); //STARTLINE 
    oled_write_command(0x8D); //CHARGEPUMP 
    oled_write_command(0x14) ;
    oled_write_command(0x20); //MEMORYMODE 
    oled_write_command(0x00); //0x0 act like ks0108 
    oled_write_command(0xA1); //SEGREMAP OR 1 
    oled_write_command(0xC8); //COMSCANDEC 
    oled_write_command(0xDA); //COMPINS 
    if(screenheight==64)oled_write_command(0x12); 
    else oled_write_command(0x02);
    oled_write_command(0x81); //SETCONTRAST 
    oled_write_command(0xCF) ;
    oled_write_command(0xd9); //SETPRECHARGE 
    oled_write_command(0xF1) ;
    oled_write_command(0xDB); //VCOMDETECT 
    oled_write_command(0x40) ;
    oled_write_command(0xA4); //DISPLAYALLON_RESUME 
    oled_write_command(0xA6); //NORMALDISPLAY 
    oled_write_command(0xAF); //DISPLAYON
    if(Option->DISPLAY_ORIENTATION&1){
        HRes=HorizontalRes;
        VRes=VerticalRes;
    } else {
        VRes=HorizontalRes;
        HRes=VerticalRes;
    }

    //Set the DrawRectangle vector to point to our function
    DrawRectangleVector= (unsigned int)&DrawRectangleMEM + libAddr;

    //Set the DrawBitmap vector to point to our function
    DrawBitmapVector=(unsigned int)&DrawBitmapMEM + libAddr;

    
    CFuncInt=(unsigned int)&updatedisplay + libAddr;
    
    //CLS
    DrawRectangle(0,0,HRes-1,VRes-1,0);


#ifndef i2c
    SPIBRG=brgsave; //restore user (or my) setup
    SPICON=consave;
    SPICON2=con2save;
    static const char startup[]="SSD1306 SPI driver loaded\r\n";
#else
    static const char startup[]="SSD1306 I2C driver loaded\r\n";
#endif
    pstring(startup);
}